#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#define BUFLEN 256
#define INPUTERR(x) {printf("Error reading input parameter \"%s\", check config file.\n",(x)); return 1;}

/* random number generation:
 * algorithm taken from usenet post by (reputable guy) George Marsaglia
 * "Random numbers for C: The END?"
 * https://groups.google.com/group/sci.math.num-analysis/msg/eb4ddde782b17051?hl=en&pli=1
 *
 * independent RNG streams are created with initstream(seed)
 */

typedef struct
{
    unsigned int z, w, jsr, jcong;
} rngstate;

unsigned int kiss(rngstate * st)
{
    st->z = 36969 * (st->z&65535) + (st->z>>16);
    st->w = 18000 * (st->w&65535) + (st->w>>16);
    st->jsr ^= (st->jsr<<17);
    st->jsr ^= (st->jsr>>13);
    st->jsr ^= (st->jsr<< 5);
    st->jcong = 69069 * st->jcong+1234567;
    return (((st->z<<16) + st->w) ^ (st->jcong)) + st->jsr;
}

double kissunif(rngstate * st) {return kiss(st) / (double) 4294967296;}

rngstate * initstream(unsigned int seed)
{
    int i;
    rngstate * st;
    st = malloc(sizeof(rngstate));
    assert(st);
    st->z = st->w = st->jsr = st->jcong = seed; 
    st->z = kiss(st);
    st->w = kiss(st);
    st->jsr = kiss(st);
    st->jcong = kiss(st);
    /* warm up the stream */
    for (i=0;i<256;i++) kiss(st);
    return(st);
}

#ifdef DEBUG
void testrng()
{
    int i;
    unsigned int k;
    rngstate * st1;
    st1 = initstream(1);
    /* note: these values taken from the Marsaglia post.
     * the 1,000,257th call to kiss()
     * from starting with this state should give 1372460312.
     * another 1,000,001 calls should produce the listed
     * subgenerator values. */
    st1->z = 12345;
    st1->w = 65435;
    st1->jsr = 34221;
    st1->jcong = 12345; 
    for(i=1;i<1000001+256;i++){k=kiss(st1) ;} 
    printf("RNG test: these should be zero: %d ",k-1372460312U);
    for(i=1;i<1000001;i++){kiss(st1) ;} 
    printf("%d %d %d\n",st1->jcong-1529210297U,st1->jsr-2642725982U,
            (st1->z<<16)+(st1->w)-904977562U);
    free(st1);
}
#endif

/* data structures:
 * we have markers and edge nodes.  A slightly confusing thing
 * is that the nodes here are actually the edges of the VLMC in, e.g.,
 * Browning 2006 AJHG.
 *
 * markers: []; edges: o
 *
 * [] -> o 
 * |
 * v
 * [] -> o -> o -> o -> o
 * |
 * v
 * [] -> o -> o 
 *
 * at the previous marker, we have selected a child I and counted the total # = N
 * of edges going into that child.  at the next marker, we choose a random integer K
 * between 1 and N and traverse the chain of edge nodes until the running
 * count of the nodes with parent = I is greater than K.  that edge gives
 * and the omitted allele.
 *
 * for a recombination jump, we follow the same procedure except we can select 
 * any edge, not just those with a specific parent.
 * 
 */

typedef struct _marker 
{
    struct _edge * edges;
    struct _marker * nextmarker;
    int index;
} marker;

typedef struct _edge 
{
    struct _edge * nextedge;
    int parent;
    int child;
    int allele;
    int count;
    int nsamechild;
} edge;


marker * addmarker(marker * currmarker, int index)
{
    marker * newmarker = malloc(sizeof(marker));
    assert(newmarker);
    currmarker->nextmarker = newmarker;
    newmarker->index = index;
    newmarker->nextmarker = NULL;
    return newmarker;
}

edge * addedge(edge * head, int parent, int child, int allele, int count)
{
    edge * newedge;
    newedge = malloc(sizeof(edge));
    assert(newedge);
    newedge->nextedge = head;
    newedge->parent = parent;
    newedge->child = child;
    newedge->allele = allele;
    newedge->count = count;
    return newedge;
}

/* check if input dag file is in .gz format and, if necessary,
 * create a decompressed copy */
int gunzipit(char * dagfilename)
{
    char * gzsuf = ".gz";
    int i;
    for(i=0;i<BUFLEN;i++) if('\0'==dagfilename[i]) break;
    if(!strncmp(dagfilename+i-3,gzsuf,3))
    {
        dagfilename[i-3] = '\0'; /* chop off .gz suffix */
        if(!access(dagfilename,R_OK))
        {
            printf("Using inflated file %s\n",dagfilename);
        } else
        {
            char gzcommand[256];
            printf("Inflating %s.gz.\nYou may want to delete %s afterwards.\n",dagfilename,dagfilename);
            sprintf(gzcommand,"gunzip -c %s.gz > %s",dagfilename,dagfilename);
            printf("%s\n",gzcommand);
            if(system(gzcommand))
            {
                printf("Error inflating %s.gz\n",dagfilename); 
                return 1;
            }
        }
    }
    return 0;
}

void progbar(int percent)
{
    int i, prog = ceil(percent*50/(double) 100);
    printf("\r[");
    for(i=0;i<prog;i++)
    {
        printf("-");
    }
    for(;i<50;i++)
    {
        printf(" ");
    }
    printf("] %3.d%%",percent);
    fflush(stdout);
}

int main(int argv, char** argc)
{
    int i,j;
    char buf[BUFLEN];
    char c;
    char dagfilename[BUFLEN],outfilename[BUFLEN];
    FILE * configfile, * dagfile, *outfile;
    int ncreate, hapcols, outlabs, outspaces;
    int tlevel, tparent, tchild, tallele, tcount;
    char tmarker[BUFLEN];
    int nhaps, currnhaps, nmarks;
    double recombprob;
    int ** outmat;
    rngstate * recombstream, * transstream;
    unsigned int seed1, seed2;
    int currstate, transpool, runcount, transrand, recombflag;
    marker * mhead;
    marker * mcurr;
    edge * ecurr, * ecurr2;
    edge * ehead;

#ifdef DEBUG
    testrng();
#endif

    configfile = fopen(argc[1],"r");
    if(!fscanf(configfile,"dag file %s ",dagfilename)) INPUTERR("dag file");
    if(!fscanf(configfile,"recomb probability %lf ",&recombprob)) INPUTERR("recomb probability");
    if(!fscanf(configfile,"create haps %d ",&ncreate)) INPUTERR("create haps"); 
    if(!fscanf(configfile,"output file %s ",outfilename)) INPUTERR("output file"); 
    if(!fscanf(configfile,"first random seed %u ",&seed1)) INPUTERR("first random seed"); 
    if(!fscanf(configfile,"second random seed %u ",&seed2)) INPUTERR("second random seed"); 
    if(!fscanf(configfile,"output haplotype columns %d ",&hapcols)) INPUTERR("output haplotype columns");
    if(!fscanf(configfile,"output labels %d ",&outlabs)) INPUTERR("output labels"); 
    if(!fscanf(configfile,"output spaces %d ",&outspaces)) INPUTERR("output spaces");

    if(outlabs) outspaces = 1;

    if(gunzipit(dagfilename)) return 1;

    if(!(dagfile = fopen(dagfilename,"r")))
    {
        printf("Error opening dag file %s\n",dagfilename); 
        return 1;
    }

    mhead = malloc(sizeof(marker));
    assert(mhead);
    mhead->index = 0;
    mcurr = mhead;

    ehead = NULL;


    fgets(buf,BUFLEN,dagfile); /* skip header */
    while(fgets(buf,BUFLEN,dagfile))
    {
        if('\n'==buf[0]) continue; /* empty line */
        /* did we finish the line?  if not scan ahead to next line */
        for(i=0;i<BUFLEN;i++) if('\n'==buf[i]) break;
        if(BUFLEN==i) while((c=getc(dagfile),'\n'!=c&&EOF!=c));
        sscanf(buf,"%d %s %d %d %d %d",&tlevel,tmarker,&tparent,&tchild,&tallele,&tcount);
        if(tlevel!=mcurr->index) 
        /* we've moved to the next marker so wrap up the current one */
        {
            mcurr->edges = ehead; /* connect the chain of edges to the marker node */
            ehead = NULL;         /* start a new edge chain for the next marker */

            /* count total haps at marker and check that it's same as other markers.
             * also calculate total count of edges which share each edge's child.
             */
            currnhaps = 0;
            for(ecurr=mcurr->edges;ecurr;ecurr=ecurr->nextedge)
            {
                currnhaps += (ecurr->count);
                ecurr->nsamechild = 0;
                /* loop over edges and add up counts of edges with same child as ecurr.
                 * not the most efficient but easy to write */
                for(ecurr2=mcurr->edges;ecurr2;ecurr2=ecurr2->nextedge)
                {
                    if(ecurr->child==ecurr2->child) ecurr->nsamechild += ecurr2->count;
                }
            }
            if(0==mcurr->index)
            {
                nhaps = currnhaps;
                printf("DAG file contains %d haplotypes.\n",nhaps);
            } else
            {
                if(nhaps!=currnhaps)
                {
                    printf("Error: marker %d contains %d haplotypes but marker 0\
contains %d haplotypes.  Check DAG file for problems.\n",mcurr->index,currnhaps,nhaps);
                    return(1);
                }
            }

            mcurr = addmarker(mcurr,tlevel);
        }
        /* add a new edge node for this edge */
        ehead = addedge(ehead,tparent,tchild,tallele,tcount);
    }
    mcurr->edges = ehead;
    nmarks = mcurr->index+1;
    printf("Finished reading DAG file.\n");
    outmat = malloc(ncreate*sizeof(int*));
    assert(outmat);
    if(!(outfile = fopen(outfilename,"w")))
    {
        printf("Error opening output file %s\n",outfilename); 
        return 1;
    }
    /* now do the actual haplotype simulation */
    recombstream = initstream(seed1);
    transstream = initstream(seed2);

    printf("Generating %d haplotypes...\n",ncreate);
    for(i=0;i<ncreate;i++)
    {
        /* simulation works like this:
         * say there are two edges with child node 1
         * with counts 10 and 20 at marker k 
         * and three edges with parent node 1
         * with counts 5 14 and 11 * at marker k+1.
         *
         * at marker k we chose an edge with child 1, so 
         * transpool = 30 (total pool of edges we can transition to).
         * we choose a random integer from 1 to 30, say transrand=12.
         * we traverse all edges at marker k+1, skipping those
         * with parent != 1.  the first edge has count 5, 
         * so the running count=runcount
         * is 5, smaller than 12.  the next edge has count 14, so the running
         * count is 19, >= to transrand, we we choose that edge.
         *
         * for a recombinant jump, we randomly choose an edge, so we set 
         * transpool = total # of haps and do not skip any edges.  
         *
         * note that edges do not have to be in any kind of order, which 
         * simplifies things 
         *
         * two random streams are used so changes can be made to, say,
         * transition process but the recombination patterns stay the same.
         *
         */
        outmat[i] = malloc(nmarks*sizeof(int));
        assert(outmat[i]);

        currstate = 0; /* start in state 0 always */
        transpool = nhaps; /* first set of edges all have same parent (0) */
        /* step through markers */
        for(mcurr=mhead;mcurr;mcurr = mcurr->nextmarker)
        {
            /* do we recombine? if so set transpool to all haps */
            recombflag = kissunif(recombstream)<recombprob;
            if(recombflag) transpool = nhaps;

            /* choose a random haplotype index out of the pool
             * of possible haplotypes we can transition to */
            transrand = ceil(transpool*kissunif(transstream));
#ifdef DEBUG
                printf("\nmarker %d\n",mcurr->index);
                printf("transpool is size %d\n", transpool);
                printf("Recombine? %s\n",recombflag ? "Yes" : "No" );
                if (recombflag)
                {
                    printf("getting hap # %d \n", transrand);
                } else
                {
                    printf("getting hap # %d with parent %d\n", transrand,currstate);
                }
                
#endif

            /* step through edges until we find that haplotype */
            runcount = 0;
            for(ecurr=mcurr->edges;ecurr;ecurr=ecurr->nextedge)
            {
                /* if recombinant, we are choosing a random edge
                 * so doesn't matter what the parent is */
#ifdef DEBUG 
                printf("This edge has parent %d and child %d and count %d and allele %d \n",ecurr->parent,ecurr->child, ecurr->count, ecurr->allele);
                if(!recombflag&&ecurr->parent!=currstate) printf("Runcount is %d\n",runcount);
#endif
                if(!recombflag&&ecurr->parent!=currstate) continue;
                runcount += ecurr->count;
#ifdef DEBUG
                printf("Runcount is %d\n",runcount);
                if(transrand<=runcount) printf("we found the hap, new child/parent is %d, allele is %d \n",ecurr->child,ecurr->allele);
#endif
                if(transrand<=runcount) break; /* found it */
            }
#ifdef DEBUG
            printf("Press ENTER to continue\n");
            getchar();
#endif
            currstate = ecurr->child;
            /* at the next marker, size of pool of possible transitions is
             * the # of haps with this node as the parent */
            transpool = ecurr->nsamechild; 
            outmat[i][mcurr->index] = ecurr->allele;
        }
#ifndef DEBUG
        progbar(ceil(100*(i+1)/(double)ncreate));
#endif
    }
    printf("\n");


    /* finally, print everything out */
    printf("Writing haplotypes to %s...\n",outfilename);
    if(hapcols)
    {
        if(outlabs)
        {
            for(i=0;i<ncreate;i++)
            {
                fprintf(outfile,"%d ",i+1);
            }
            fprintf(outfile,"\n");
        }
        for(j=0;j<nmarks;j++)
        {
            for(i=0;i<ncreate;i++)
            {
                fprintf(outfile,"%d",outmat[i][j]);
                if(outspaces) printf(" ");
            }
            progbar(ceil(100*(j+1)/(double)nmarks));
            fprintf(outfile,"\n");
        }
    } else
    {
        for(i=0;i<ncreate;i++)
        {
            if(outlabs) fprintf(outfile,"%d ",i+1);
            for(j=0;j<nmarks;j++)
            {
                fprintf(outfile,"%d",outmat[i][j]);
                if(outspaces) printf(" ");
            }
            progbar(ceil(100*(i+1)/(double)ncreate));
            fprintf(outfile,"\n");
        }
    } 
    printf("\n");

    return 0; 
}



#if 0
// USE TO CHECK-- print out DAG file (rows in reverse order within markers)
    for(mcurr=mhead;mcurr;mcurr = mcurr->nextmarker)
    {
        for(ecurr=mcurr->edges;ecurr;ecurr=ecurr->nextedge)
        printf("%d %d %d %d %d %d\n",mcurr->index,ecurr->parent,ecurr->child,ecurr->allele,ecurr->count,ecurr->nsamechild);
    }
#endif
